"""
Collected rules from RFC 7231, Appendix c, D.
https://tools.ietf.org/html/rfc7231#appendix-C
"""

from abnf.parser import Rule as _Rule
from . import rfc4647, rfc5322, rfc5646, rfc7230
from .misc import load_grammar_rules


@load_grammar_rules(
    [
        ("language-range", rfc4647.Rule("language-range")),
        ("mailbox", rfc5322.Rule("mailbox")),
        ("language-tag", rfc5646.Rule("language-tag")),
        ("BWS", rfc7230.Rule("BWS")),
        ("OWS", rfc7230.Rule("OWS")),
        ("RWS", rfc7230.Rule("RWS")),
        ("URI-reference", rfc7230.Rule("URI-reference")),
        ("absolute-URI", rfc7230.Rule("absolute-URI")),
        ("comment", rfc7230.Rule("comment")),
        # see https://www.rfc-editor.org/errata/eid4225 for field-name.
        ("field-name", rfc7230.Rule("field-name")),
        ("partial-URI", rfc7230.Rule("partial-URI")),
        ("quoted-string", rfc7230.Rule("quoted-string")),
        ("token", rfc7230.Rule("token")),
    ]
)
class Rule(_Rule):
    """Rules from RFC 7231."""

    grammar = [
        'Accept = [ ( "," / ( media-range [ accept-params ] ) ) *( OWS "," [ OWS ( media-range [ accept-params ] ) ] ) ]',
        'Accept-Charset = *( "," OWS ) ( ( charset / "*" ) [ weight ] ) *( OWS "," [ OWS ( ( charset / "*" ) [ weight ] ) ] )',
        'Accept-Encoding = [ ( "," / ( codings [ weight ] ) ) *( OWS "," [ OWS ( codings [ weight ] ) ] ) ]',
        'Accept-Language = *( "," OWS ) ( language-range [ weight ] ) *( OWS "," [ OWS ( language-range [ weight ] ) ] )',
        'Allow = [ ( "," / method ) *( OWS "," [ OWS method ] ) ]',
        # BWS = <BWS, see [RFC7230], Section 3.2.3>
        'Content-Encoding = *( "," OWS ) content-coding *( OWS "," [ OWS content-coding ] )',
        'Content-Language = *( "," OWS ) language-tag *( OWS "," [ OWS language-tag ] )',
        "Content-Location = absolute-URI / partial-URI",
        "Content-Type = media-type",
        "Date = HTTP-date",
        'Expect = "100-continue"',
        "From = mailbox",
        "GMT = %x47.4D.54 ",
        "HTTP-date = IMF-fixdate / obs-date",
        'IMF-fixdate = day-name "," SP date1 SP time-of-day SP GMT',
        "Location = URI-reference",
        "Max-Forwards = 1*DIGIT",
        # OWS = <OWS, see [RFC7230], Section 3.2.3>
        # RWS = <RWS, see [RFC7230], Section 3.2.3>
        "Referer = absolute-URI / partial-URI",
        "Retry-After = HTTP-date / delay-seconds",
        "Server = product *( RWS ( product / comment ) )",
        # URI-reference = <URI-reference, see [RFC7230], Section 2.7>
        "User-Agent = product *( RWS ( product / comment ) )",
        'Vary = "*" / ( *( "," OWS ) field-name *( OWS "," [ OWS field-name ] ) )',
        # absolute-URI = <absolute-URI, see [RFC7230], Section 2.7>
        'accept-ext = OWS ";" OWS token [ "=" ( token / quoted-string ) ]',
        "accept-params = weight *accept-ext",
        "asctime-date = day-name SP date3 SP time-of-day SP year",
        "charset = token",
        'codings = content-coding / "identity" / "*"',
        # comment = <comment, see [RFC7230], Section 3.2.6>
        "content-coding = token",
        "date1 = day SP month SP year",
        'date2 = day "-" month "-" 2DIGIT',
        "date3 = month SP ( 2DIGIT / ( SP DIGIT ) )",
        "day = 2DIGIT",
        "day-name = %x4D.6F.6E / %x54.75.65 / %x57.65.64 / %x54.68.75 / %x46.72.69 / %x53.61.74 / %x53.75.6E ",
        "day-name-l = %x4D.6F.6E.64.61.79 / %x54.75.65.73.64.61.79 / %x57.65.64.6E.65.73.64.61.79 / %x54.68.75.72.73.64.61.79 / %x46.72.69.64.61.79 / %x53.61.74.75.72.64.61.79 / %x53.75.6E.64.61.79 ",
        "delay-seconds = 1*DIGIT",
        # field-name = <comment, see [RFC7230], Section 3.2>
        "hour = 2DIGIT",
        # language-range = <language-range, see [RFC4647], Section 2.1>
        # language-tag = <Language-Tag, see [RFC5646], Section 2.1>',
        # mailbox = <mailbox, see [RFC5322], Section 3.4>
        'media-range = ( "*/*" / ( type "/*" ) / ( type "/" subtype ) ) *( OWS ";" OWS parameter )',
        'media-type = type "/" subtype *( OWS ";" OWS parameter )',
        "method = token",
        "minute = 2DIGIT",
        "month = %x4A.61.6E / %x46.65.62 / %x4D.61.72 / %x41.70.72 / %x4D.61.79 / %x4A.75.6E / %x4A.75.6C / %x41.75.67 / %x53.65.70 / %x4F.63.74 / %x4E.6F.76 / %x44.65.63 ",
        "obs-date = rfc850-date / asctime-date",
        'parameter = token "=" ( token / quoted-string )',
        # partial-URI = <partial-URI, see [RFC7230], Section 2.7>
        'product = token [ "/" product-version ]',
        "product-version = token",
        # quoted-string = <quoted-string, see [RFC7230], Section 3.2.6>
        'qvalue = ( "0" [ "." *3DIGIT ] ) / ( "1" [ "." *3"0" ] )',
        'rfc850-date = day-name-l "," SP date2 SP time-of-day SP GMT',
        "second = 2DIGIT",
        "subtype = token",
        'time-of-day = hour ":" minute ":" second',
        # token = <token, see [RFC7230], Section 3.2.6>
        "type = token",
        'weight = OWS ";" OWS "q=" qvalue',
        "year = 4DIGIT",
    ]



"""
Collected rules from RFC 7232
https://tools.ietf.org/html/rfc7232
"""

from abnf.parser import Rule as _Rule
from . import rfc7230, rfc7231
from .misc import load_grammar_rules


@load_grammar_rules(
    [
        ("HTTP-date", rfc7231.Rule("HTTP-date")),
        ("OWS", rfc7230.Rule("OWS")),
        ("obs-text", rfc7230.Rule("obs-text")),
    ]
)
class Rule(_Rule):
    """Rules from RFC 7232."""

    grammar = [
        "ETag = entity-tag",
        # HTTP-date = <HTTP-date, see [RFC7231], Section 7.1.1.1>
        'If-Match = "*" / ( *( "," OWS ) entity-tag *( OWS "," [ OWS entity-tag ] ) )',
        "If-Modified-Since = HTTP-date",
        'If-None-Match = "*" / ( *( "," OWS ) entity-tag *( OWS "," [ OWS entity-tag ] ) )',
        "If-Unmodified-Since = HTTP-date",
        "Last-Modified = HTTP-date",
        # OWS = <OWS, see [RFC7230], Section 3.2.3>
        "entity-tag = [ weak ] opaque-tag",
        'etagc = "!" / %x23-7E / obs-text',
        # obs-text = <obs-text, see [RFC7230], Section 3.2.6>
        "opaque-tag = DQUOTE *etagc DQUOTE",
        "weak = %x57.2F",
    ]



"""
Collected rules from RFC 7235
https://tools.ietf.org/html/rfc7235
"""

from abnf.parser import Rule as _Rule
from . import rfc7230
from .misc import load_grammar_rules


@load_grammar_rules(
    [
        ("BWS", rfc7230.Rule("BWS")),
        ("OWS", rfc7230.Rule("OWS")),
        ("token", rfc7230.Rule("token")),
        ("quoted-string", rfc7230.Rule("quoted-string")),
    ]
)
class Rule(_Rule):
    """Rules from RFC 7235."""

    grammar = [
        "Authorization = credentials",
        # BWS = <BWS, see [RFC7230], Section 3.2.3>',
        # OWS = <OWS, see [RFC7230], Section 3.2.3>',
        # See discussion below for WWW-Authenticate.
        #'Proxy-Authenticate = *( "," OWS ) challenge *( OWS "," [ OWS challenge ] )',
        'Proxy-Authenticate = *( "," OWS ) challenge *( OWS ("," / challenge) )',
        "Proxy-Authorization = credentials",
        # The rule WWW-Authenticate is ambiguous. I am not the first person to observe this
        # <https://www.ietf.org/mail-archive/web/httpbisa/current/msg07914.html>.
        # The problem is essentially nested comma-separated lists.  Using the rules given
        # in RFC 7235 with this parser, parsing source 'Basic realm="foo", Pascal realm="bar"
        # using the rule WWW-Authenticate will result in the consumption of 'Basic realm="foo",'.
        # So we tinker a bit to account for the challenge rule consuming the trailing comma.
        #'WWW-Authenticate = *( "," OWS ) challenge *( OWS "," [ OWS challenge ] )',
        'WWW-Authenticate = *( "," OWS ) challenge *( OWS ("," / challenge) )',
        'auth-param = token BWS "=" BWS ( token / quoted-string )',
        "auth-scheme = token",
        'challenge = auth-scheme [ 1*SP ( token68 / [ ( "," / auth-param ) *(OWS "," [ OWS auth-param ] ) ] ) ]',
        'credentials = auth-scheme [ 1*SP ( token68 / [ ( "," / auth-param ) *( OWS "," [ OWS auth-param ] ) ] ) ]',
        # quoted-string = <quoted-string, see [RFC7230], Section 3.2.6>',
        # token = <token, see [RFC7230], Section 3.2.6>',
        'token68 = 1*( ALPHA / DIGIT / "-" / "." / "_" / "~" / "+" / "/" ) *"="',
    ]



"""
This is the ABNF grammar, expressed in ABNF, plus the core rules.

Collected rules from RFC 5234
https://tools.ietf.org/html/rfc5234
"""

from abnf.parser import Rule as _Rule
from .misc import load_grammar_rules


@load_grammar_rules()
class Rule(_Rule):
    """Rule objects generated from ABNF in RFC 5234."""

    grammar = [
        "rulelist = 1*( rule / (*c-wsp c-nl) )",
        "rule = rulename defined-as elements c-nl\
                                        ; continues if next line starts\
                                        ;  with white space",
        'rulename = ALPHA *(ALPHA / DIGIT / "-")',
        'defined-as = *c-wsp ("=" / "=/") *c-wsp\
                                        ; basic rules definition and\
                                        ;  incremental alternatives',
        "elements = alternation *c-wsp",
        "c-wsp = WSP / (c-nl WSP)",
        "c-nl = comment / CRLF\
                                        ; comment or newline",
        'comment = ";" *(WSP / VCHAR) CRLF',
        'alternation = concatenation\
                                   *(*c-wsp "/" *c-wsp concatenation)',
        "concatenation = repetition *(1*c-wsp repetition)",
        "repetition = [repeat] element",
        'repeat = 1*DIGIT / (*DIGIT "*" *DIGIT)',
        "element = rulename / group / option /\
                                   char-val / num-val",
        'group = "(" *c-wsp alternation *c-wsp ")"',
        'option = "[" *c-wsp alternation *c-wsp "]"',
        "char-val = DQUOTE *(%x20-21 / %x23-7E) DQUOTE\
                                        ; quoted string of SP and VCHAR\
                                        ;  without DQUOTE",
        'num-val = "%" (bin-val / dec-val / hex-val)',
        'bin-val = "b" 1*BIT\
                                   [ 1*("." 1*BIT) / ("-" 1*BIT) ]\
                                        ; series of concatenated bit values\
                                        ;  or single ONEOF range',
        'dec-val = "d" 1*DIGIT\
                                   [ 1*("." 1*DIGIT) / ("-" 1*DIGIT) ]',
        'hex-val = "x" 1*HEXDIG\
                                   [ 1*("." 1*HEXDIG) / ("-" 1*HEXDIG) ]\
                                        ; white space',
    ]



import typing

from abnf.parser import Rule as _Rule, Literal
from abnf.grammars.misc import load_grammar_rules


class ImportRule(_Rule):
    pass
    
ImportRule('test', Literal('test'))    


@load_grammar_rules(
[
('test', ImportRule('test'))
])
class Rule(_Rule):
    """Rules from RFC 5646."""

    grammar: typing.List[str] = []

def test_misc_load_grammar_rules_import():
    assert Rule('test').definition == ImportRule('test').definition





"""
Collected rules from RFC 9116
https://www.rfc-editor.org/rfc/rfc9116
"""


from abnf.parser import Rule as _Rule
from . import rfc3986, rfc5322, rfc3629, rfc5646, rfc3339
from .misc import load_grammar_rules



@load_grammar_rules(
    [
    ("email", rfc5322.Rule("address")),
    ("uri", rfc3986.Rule("uri")),
    ('field-name', rfc5322.Rule("field-name")),
    ('unstructured', rfc5322.Rule('unstructured')),
    ('UTF8-octets', rfc3629.Rule('UTF8-octets')),
    ('UTF8-char', rfc3629.Rule('UTF8-char')),
    ('UTF8-1', rfc3629.Rule('UTF8-1')),
    ('UTF8-2', rfc3629.Rule('UTF8-2')),
    ('UTF8-3', rfc3629.Rule('UTF8-3')),
    ('UTF8-4', rfc3629.Rule('UTF8-4')),
    ('UTF8-tail', rfc3629.Rule('UTF8-tail')),
    ('lang-tag', rfc5646.Rule('langtag')),
    ('date-time', rfc3339.Rule('date-time')),
    ]
)
class Rule(_Rule):
    """Rules from RFC 5987."""

    grammar = [
        'body =  signed / unsigned',

        'unsigned =  *line (contact-field eol) *line (expires-field eol) *line [lang-field eol] *line',
        

        #'; signed is the production that should match the OpenPGP clearsigned',
        #'; document',
        'signed =  cleartext-header 1*(hash-header) CRLF cleartext signature',

        'cleartext-header =  %s"-----BEGIN PGP SIGNED MESSAGE-----" CRLF',

        'hash-header      =  %s"Hash: " hash-alg *("," hash-alg) CRLF',

        'hash-alg         =  token',

        'cleartext        =  *((line-dash / line-from / line-nodash) [CR] LF)',

        'line-dash        =  ("- ") "-" *UTF8-char-not-cr',

        'line-from        =  ["- "] "From " *UTF8-char-not-cr',

        'line-nodash      =  ["- "] *UTF8-char-not-cr',

        'UTF8-char-not-dash =  UTF8-1-not-dash / UTF8-2 / UTF8-3 / UTF8-4',
        'UTF8-1-not-dash  =  %x00-2C / %x2E-7F',
        'UTF8-char-not-cr =  UTF8-1-not-cr / UTF8-2 / UTF8-3 / UTF8-4',
        'UTF8-1-not-cr    =  %x00-0C / %x0E-7F',

        #'; UTF8 rules from RFC 3629 -- imported above',


        'signature        =  armor-header armor-keys CRLF signature-data armor-tail',

        'armor-header     =  %s"-----BEGIN PGP SIGNATURE-----" CRLF',

        'armor-keys       =  *(token ": " *( VCHAR / WSP ) CRLF)',

        'armor-tail       =  %s"-----END PGP SIGNATURE-----" CRLF',

        'signature-data   =  1*(1*(ALPHA / DIGIT / "=" / "+" / "/") CRLF)\
                            ; base64; see RFC 4648\
                            ; includes RFC 4880 checksum',

        'line             =  [ (field / comment) ] eol',

        'eol              =  *WSP [CR] LF',

        #'field            =  ; optional fields\
        'field            =  ack-field / can-field / contact-field / encryption-field / hiring-field / policy-field / ext-field',

        'fs               =  ":"',

        'comment          =  "#" *(WSP / VCHAR / %x80-FFFFF)',

        'ack-field        =  "Acknowledgments" fs SP uri',

        'can-field        =  "Canonical" fs SP uri',

        'contact-field    =  "Contact" fs SP uri',

        'expires-field    =  "Expires" fs SP date-time',

        'encryption-field =  "Encryption" fs SP uri',

        'hiring-field     =  "Hiring" fs SP uri',

        'lang-field       =  "Preferred-Languages" fs SP lang-values',

        'policy-field     =  "Policy" fs SP uri',

        #date-time        =  < imported from Section 5.6 of [RFC3339] >

        #lang-tag         =  < Language-Tag from Section 2.1 of [RFC5646] >

        'lang-values      =  lang-tag *(*WSP "," *WSP lang-tag)',

        #uri              =  < URI as per Section 3 of [RFC3986] >

        'ext-field        =  field-name fs SP unstructured',

        #field-name       =  < imported from Section 3.6.8 of [RFC5322] >

        #unstructured     =  < imported from Section 3.2.5 of [RFC5322] >

        #token            =  < imported from Section 5.1 of [RFC2045] >
        # definition from RFC 2045:
        # token := 1*<any (US-ASCII) CHAR except SPACE, CTLs,
        #        or tspecials>

        # tspecials :=  "(" / ")" / "<" / ">" / "@" /
        #           "," / ";" / ":" / "\" / <">
        #           "/" / "[" / "]" / "?" / "="
        #          ; Must be in quoted-string,
        #          ; to use within parameter values

        
        'token-char = %x21-27 / %x2A-2B / %x30-39 / %x41-5A / %x5E-7E',
        'token = 1*token-char',


# core rules included in this grammar are omitted.

    ]



import pytest

from abnf.grammars import rfc9116



@pytest.mark.parametrize("src", [
'''-----BEGIN PGP SIGNED MESSAGE-----\r\nHash: SHA512\r\n\r\nContact: charles@declaresub.com\r\nExpires: 2023-03-14T00:00:00.000Z\r\n-----BEGIN PGP SIGNATURE-----\r\n\r\niHUEARYKAB0WIQSsP2kEdoKDVFpSg6u3rK+YCkjapwUCYhjpQwAKCRC3rK+YCkjapyk2AP97ePaFUo8K8e1D+N+G6caqXjC/pwnZB+Wkk15AI+xstgD/VR5rOLKLZ7QFgKk5ohVS7qHou8Ux9cdodY2BRUIdrww==gFfQ\r\n-----END PGP SIGNATURE-----\r\n''',
])
def test_securitytxt_contact(src: str):
    assert rfc9116.Rule('body').parse_all(src)



import pytest

from abnf.grammars import rfc7489
from abnf.parser import Source


def test_valid_dmarc_version():
    version = rfc7489.Rule('dmarc-version')
    assert version.parse_all("v=DMARC1")
    assert version.parse_all("v =  DMARC1")
    assert version.parse_all("V =  DMARC1")


@pytest.mark.parametrize("src", [
    'mailto:olivier@hureau.com',
    'mailto:olivier@hureau.com!50m',
    'mailto:olivier@hureau.com!50',
    'mailto:olivier@hureau.com!50'
])

def test_valid_dmarc_uri(src: Source):
    uri = rfc7489.Rule('dmarc-uri')
    assert uri.parse_all(src)

@pytest.mark.parametrize("src", [
    ';',
    ' ;',
    '; ',
    '\t;'
])

def test_valid_dmarc_seperator(src: Source):
    separator = rfc7489.Rule('dmarc-sep')
    assert separator.parse_all(src)

@pytest.mark.parametrize("src", [
    'p=none', 'p=reject', 'p=quarantine',
    'P=none', 'p=NoNe',
    'p =    none'
])

def test_valid_dmarc_request(src: Source):
    request = rfc7489.Rule('dmarc-request')
    assert request.parse_all(src)

@pytest.mark.parametrize("src", [
    'sp=none', 'sp=reject', 'sp=quarantine',
    'sP=none', 'Sp=NoNe',
    'sp =    none'
])

def test_valid_dmarc_srequest(src: Source):
    srequest = rfc7489.Rule('dmarc-srequest')
    assert srequest.parse_all(src)

@pytest.mark.parametrize("src", [
    'rua=mailto:olivier@hureau.com',
    'RUA=mailto:olivier@hureau.com!50m',
    'rua  =   mailto:olivier@hureau.com!50',
    'rua=mailto:olivier@hureau.com, mailto:olivier@hureau.com',
    'rua=  mailto:olivier@hureau.com   , mailto:olivier@hureau.com   '
])

def test_valid_dmarc_auri(src: Source):
    auri = rfc7489.Rule('dmarc-auri')
    assert auri.parse_all(src)

@pytest.mark.parametrize("src", [
    'ruf=mailto:olivier@hureau.com',
    'RUF=mailto:olivier@hureau.com!50m',
    'ruf  =   mailto:olivier@hureau.com!50',
    'ruf=mailto:olivier@hureau.com, mailto:olivier@hureau.com',
    'ruf=  mailto:olivier@hureau.com   , mailto:olivier@hureau.com   '
])

def test_valid_dmarc_furi(src: Source):
    auri = rfc7489.Rule('dmarc-furi')
    assert auri.parse_all(src)

@pytest.mark.parametrize("src", [
    'aspf=s',
    'aspf=r',
    'aspf=R',
    'aspf=S',
    'aspf = s',
    'aSPf = s',
])

def test_valid_dmarc_aspf(src: Source):
    aspf = rfc7489.Rule('dmarc-aspf')
    assert aspf.parse_all(src)

@pytest.mark.parametrize("src", [
    'ri=8600',
    'ri  = 8600',
    'RI = 8600'
])

def test_valid_dmarc_ri(src: Source):
    interval = rfc7489.Rule('dmarc-ainterval')
    assert interval.parse_all(src)

@pytest.mark.parametrize("src", [
    'fo=1',
    'fo=  1',
    'fo=1:d:s',
    'fo=1 :  d: s',
    'FO=1',
    'fo=1:1:1:1', # This one should not pass but abnf is abnf...
])

def test_valid_dmarc_fo(src: Source):
    fo = rfc7489.Rule('dmarc-fo')
    assert fo.parse_all(src)

@pytest.mark.parametrize("src", [
    'rf=afrf',
    'RF=afrf',
    'rf  =  afrf',
    'rf=aFrF',
])

def test_valid_dmarc_rf(src: Source):
    rf = rfc7489.Rule('dmarc-rfmt')
    assert rf.parse_all(src)

@pytest.mark.parametrize("src", [
    'pct=100',
    'pct=99',
    'pct=5',
    'pct=0',
    'PCT=0',
    'PCT  =  0'
])

def test_valid_dmarc_pct(src: Source):
    pct = rfc7489.Rule('dmarc-percent')
    assert pct.parse_all(src)

@pytest.mark.parametrize("src", [
    'v=DMARC1;p=reject;',
    'v=DMARC1;p=reject;sp=quarantine;rua=mailto:olivier@hureau.com;ruf=mailto:olivier@hureau.com;adkim=s;aspf=s;ri=2400;fo=1:d:s;rf=afrf;pct=0;'
])

def test_valid_dmarc_record(src: Source):
    record = rfc7489.Rule('dmarc-record')
    assert record.parse_all(src)






"""Conditional typing imports are collected here, because doing the import below in 
parser.py appears to confuse pylance, causing it to flag objects as not conforming to
Parser protocol.  Adding Parser as a superclass of Rule leads to more odd type errors.
The code branches on version instead of using try-except so that mypy is happy.
"""

import sys # pragma: no cover

if sys.version_info >= (3, 8): # pragma: no cover
    from typing import Protocol 
else:
    from typing_extensions import Protocol # pragma: no cover

__all__ = ["Protocol"]  # pragma: no cover



import pytest

from abnf.grammars import rfc7235

@pytest.mark.parametrize("src", [
    'Basic bW9vZgo=',
    'Digest username="Mufasa", realm="http-auth@example.org", uri="/dir/index.html", algorithm=MD5, nonce="7ypf/xlj9XXwfDPEoM4URrv/xwf94BcCAzFZH4GiTo0v", nc=00000001, cnonce="f2/wE4q74E6zIJEtWaHKaf5wv/H5QzzpXusqGemxURZJ", qop=auth, response="8ca523f5e9506fed4657c9700eebdbec", opaque="FQhe/qaU925kfnzjCev0ciny7QMkPqMAFRtzCUYo5tdS"'
    ])
def test_rfc7235_credentials(src: str):
    assert rfc7235.Rule('Authorization').parse_all(src)

@pytest.mark.parametrize("src", [
    'Basic realm="/"',
    'Basic bW9vZgo=',
    'Newauth realm="apps", type=1, title="Login to \\"apps\\"", Basic realm="simple"'
    ])    
def test_rfc7235_www_authenticate(src: str):
    assert rfc7235.Rule('WWW-Authenticate').parse_all(src)







"""
Collected rules from RFC 3339, Appendix A.
https://datatracker.ietf.org/doc/html/rfc3339
"""

from abnf.parser import Rule as _Rule
from .misc import load_grammar_rules


@load_grammar_rules()
class Rule(_Rule):
    """Rules from RFC 3339."""

    grammar = [
        'date-fullyear   = 4DIGIT',
        'date-month      = 2DIGIT  ; 01-12',
        'date-mday       = 2DIGIT  ; 01-28, 01-29, 01-30, 01-31 based on\
                                    ; month/year',
        'time-hour       = 2DIGIT  ; 00-23',
        'time-minute     = 2DIGIT  ; 00-59',
        'time-second     = 2DIGIT  ; 00-58, 00-59, 00-60 based on leap second\
                                    ; rules',
        'time-secfrac    = "." 1*DIGIT',
        'time-numoffset  = ("+" / "-") time-hour ":" time-minute',
        'time-offset     = "Z" / time-numoffset',

        'partial-time    = time-hour ":" time-minute ":" time-second\
                            [time-secfrac]',
        'full-date       = date-fullyear "-" date-month "-" date-mday',
        'full-time       = partial-time time-offset',

        'date-time       = full-date "T" full-time',
    ]



"""
Collected rules from RFC 3986, Appendix A.
https://tools.ietf.org/html/rfc3986#appendix-A
"""

from abnf.parser import Rule as _Rule
from .misc import load_grammar_rules


@load_grammar_rules()
class Rule(_Rule):
    """Rules from RFC 3986."""

    grammar = [
        'URI = scheme ":" hier-part [ "?" query ] [ "#" fragment ]',
        'hier-part = "//" authority path-abempty / path-absolute / path-rootless / path-empty',
        "URI-reference = URI / relative-ref",
        'absolute-URI = scheme ":" hier-part [ "?" query ]',
        'relative-ref = relative-part [ "?" query ] [ "#" fragment ]',
        'relative-part = "//" authority path-abempty / path-absolute / path-noscheme / path-empty',
        'scheme = ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )',
        'authority = [ userinfo "@" ] host [ ":" port ]',
        'userinfo = *( unreserved / pct-encoded / sub-delims / ":" )',
        "host = IP-literal / IPv4address / reg-name",
        "port = *DIGIT",
        'IP-literal = "[" ( IPv6address / IPvFuture ) "]"',
        'IPvFuture = "v" 1*HEXDIG "." 1*( unreserved / sub-delims / ":" )',
        'IPv6address = 6( h16 ":" ) ls32 / "::" 5( h16 ":" ) ls32 / [ h16 ] "::" 4( h16 ":" ) ls32 / [ *1( h16 ":" ) h16 ] "::" 3( h16 ":" ) ls32 / [ *2( h16 ":" ) h16 ] "::" 2( h16 ":" ) ls32 / [ *3( h16 ":" ) h16 ] "::" h16 ":" ls32 / [ *4( h16 ":" ) h16 ] "::" ls32 / [ *5( h16 ":" ) h16 ] "::" h16 / [ *6( h16 ":" ) h16 ] "::"',
        "h16 = 1*4HEXDIG",
        'ls32 = ( h16 ":" h16 ) / IPv4address',
        'IPv4address = dec-octet "." dec-octet "." dec-octet "." dec-octet',
        # alternatives reordered for correct matching.
        'dec-octet = "25" %x30-35 / "2" %x30-34 DIGIT / "1" 2DIGIT / %x31-39 DIGIT / DIGIT',
        "reg-name = *( unreserved / pct-encoded / sub-delims )",
        "path = path-abempty / path-absolute / path-noscheme / path-rootless / path-empty",
        'path-abempty = *( "/" segment )',
        'path-absolute = "/" [ segment-nz *( "/" segment ) ]',
        'path-noscheme = segment-nz-nc *( "/" segment )',
        'path-rootless = segment-nz *( "/" segment )',
        "path-empty = 0pchar",
        "segment = *pchar",
        "segment-nz = 1*pchar",
        'segment-nz-nc = 1*( unreserved / pct-encoded / sub-delims / "@" )',
        'pchar = unreserved / pct-encoded / sub-delims / ":" / "@"',
        'query = *( pchar / "/" / "?" )',
        'fragment = *( pchar / "/" / "?" )',
        'pct-encoded = "%" HEXDIG HEXDIG',
        'unreserved = ALPHA / DIGIT / "-" / "." / "_" / "~"',
        "reserved = gen-delims / sub-delims",
        'gen-delims = ":" / "/" / "?" / "#" / "[" / "]" / "@"',
        'sub-delims = "!" / "$" / "&" / "\'" / "(" / ")" / "*" / "+" / "," / ";" / "="',
    ]



import pathlib
from typing import Any, cast, Tuple

import pytest

from abnf.parser import *




def test_sorted_by_longest_match():
    match0 = Match([], 0)
    match1 = Match([], 1)
    match2 = Match([], 2)
    assert sorted_by_longest_match([match0, match1, match2]) == [match2, match1, match0]


def test_next_longest():
    match0 = Match([], 0)
    match1 = Match([], 1)
    match2 = Match([], 2)
    assert [x for x in next_longest(set([match0, match1, match2]))] == [match2, match1, match0]


def test_match_str():
    match = Match([], 0)
    assert str(match)


def test_oarse_cache_bad_max_size():
    with pytest.raises(ValueError):
        ParseCache(-12)

def test_oarse_cache_miss():
    cache = ParseCache()
    with pytest.raises(KeyError):
        cache[('foo', 1)]
    assert cache.misses == 1

def test_cache_hit():
    cache = ParseCache()
    match_set = set([ Match([], 0)])
    cache[('foo', 1)] = match_set
    assert len(cache) == 1
    assert cache[('foo', 1)] == match_set
    assert cache.hits == 1
    assert cache.misses == 0

def test_parse_cache_max_size():
    cache = ParseCache(max_size=1)
    cache[('foo', 1)] = set([ Match([], 0)])
    assert len(cache) == 1
    match_set = set([ Match([], 2)])
    cache[('foo', 3)] = match_set
    assert len(cache) == 1
    assert cache[('foo', 3)] == match_set

def test_parse_cache_del_item():
    cache = ParseCache(max_size=1)
    cache[('foo', 1)] = set([ Match([], 0)])
    assert len(cache) == 1
    del cache[('foo', 1)]
    assert len(cache) == 0
    cache = ParseCache()
    cache[('foo', 1)] = set([ Match([], 0)])

def test_parse_cache_iter():
    cache = ParseCache()
    cache[('foo', 1)] = set([ Match([], 0)])
    assert [key for key in cache] == [('foo', 1)]

def test_parse_cache_eq():
    cache = ParseCache()
    assert cache == cache

def test_parse_cache_not_eq():
    cache1 = ParseCache()
    cache2 = ParseCache()
    assert cache1 != cache2

def test_parse_cache_str():
    assert str(ParseCache())

def test_parse_cache_clear_caches():
    cache = ParseCache()
    cache[('foo', 1)] = set([ Match([], 0)])
    cache[('foo', 1)] # trigger a hit
    try:
        cache[('bar', 1)] # trigger a miss
    except KeyError:
        pass

    ParseCache.clear_caches()
    for c in ParseCache.list():
        assert len(c) == 0
        assert c.misses == 0
        assert c.hits == 0



@pytest.mark.parametrize("args", [(None, 1), (Literal('a'), None)])
def test_parseerror_bad_args(args: Tuple[Any, Any]):
    with pytest.raises(ValueError):
        ParseError(*args)

def test_parseerror_str():
    # I'm not checking the output, just exercising ParseError.__str__ .
    assert str(ParseError(Literal('a'), 1))



def test_backtracking():
    src = 'aababb'
    parser = Concatenation(Repetition(Repeat(), Alternation(Literal('a'), Literal('b'))), Literal('b'))
    assert len(parser.lparse_cache) == 0
    result = parser.lparse(src, 0)
    match = next(result)
    assert ''.join(n.value for n in match.nodes) == src


def test_alternation_first_match():
    parser = Alternation(Literal('a'), Literal('ab'), first_match=True)
    result = parser.lparse('ab', 0)
    match = next(result)
    assert ''.join(n.value for n in match.nodes) == 'a'
    assert match.start == 1

def test_alternation_fail():
    parser = Alternation(Literal('a'), Literal('b'))
    result = parser.lparse('c', 0)
    with pytest.raises(ParseError):
        next(result)


def test_alternation_str():
    assert str(Alternation(Literal('a'), Literal('b')))

def test_concatenation_from_cache():
    parser = Concatenation(Literal('a'))
    result = parser.lparse('a', 0)
    next(result) # we need to run the generator to get parser to cache result.
    cached_matchset = parser.lparse_cache[('a', 0)]
    assert isinstance(cached_matchset, set)
    cached_match = [x for x in cached_matchset][0]
    cached_match.from_cache = True # type: ignore
    result = parser.lparse('a', 0)
    match = next(result)
    assert hasattr(match, 'from_cache')

def test_concatenation_from_cache_parse_error():
    parser = Concatenation(Literal('a'))
    result = parser.lparse('b', 0)
    try:
        next(result) # we need to run the generator to get parser to cache result.
    except ParseError as exc:
        pass
    cached_error = parser.lparse_cache[('b', 0)]
    assert isinstance(cached_error, ParseError)
    result = parser.lparse('b', 0)
    try:
        next(result)
    except ParseError as exc:
        assert exc is cached_error

def test_concatenation_str():
    assert str(Concatenation(Literal('a')))


def test_repeat_str():
    assert str(Repeat())

def test_repetition_str():
    assert str(Repetition(Repeat(1, 2), Literal('foo')))

def test_option_str():
    assert str(Option(Literal('foo')))

def test_literal():
    parser = Literal("a")
    matches = parser.lparse("a", 0)
    match = next(matches)
    assert match == Match([cast(Node, LiteralNode("a", 0, 1))], 1)


def test_literal_bad_arg():
    with pytest.raises(TypeError):
        Literal(('x', 1)) # type: ignore

class FMARule(Rule):
    pass

def test_rule_first_match_alternation_get():
    rule = FMARule("test-first-match-alternation")
    assert rule.first_match_alternation is False
    rule.definition = Alternation(Literal("a"), Literal("ab"))
    assert rule.first_match_alternation is False

def test_rule_first_match_alternation_set():
    rule = FMARule("test-first-match-alternation")
    assert rule.first_match_alternation is False
    node, start = rule.parse("ab", 0)
    assert node.value == 'ab'
    assert start == 2
    rule.first_match_alternation = True
    assert rule.first_match_alternation is True
    node, start = rule.parse("ab", 0)
    assert node.value == 'a'
    assert start == 1
    rule.first_match_alternation = False
    assert rule.first_match_alternation is False
    node, start = rule.parse("ab", 0)
    assert node.value == 'ab'
    assert start == 2

def test_rule_first_match_alternation_grammar_error():
    rule = FMARule("no-grammar")
    with pytest.raises(GrammarError):
        rule.first_match_alternation = False

def test_rule_first_match_alternation_pass():
    # this test just exercises first_match_alternation setter
    # in the case of a rule which is not an alternation.
    rule = FMARule('no-alternation', Literal('a'))
    rule.first_match_alternation = True


def test_rule_grammar_error():
    with pytest.raises(GrammarError):
        FMARule('no-definition').parse('a', 0)

class ExcludeRule(Rule):
    pass
ExcludeRule.create('foo = %x66.6f.6f')
ExcludeRule.create('keyword = foo')
ExcludeRule.create('identifier = ALPHA *(ALPHA / DIGIT )')

def test_exclusion():
    identifier = ExcludeRule('identifier')
    identifier.exclude_rule(ExcludeRule('keyword'))
    assert identifier.parse_all('foobar')
    with pytest.raises(ParseError):
        identifier.parse_all('foo') # this will match 'fo', perhaps because of backtrackery.
        # so we want a different test to trigger the parse error.

def test_exclusion_2():
    identifier = ExcludeRule('identifier')
    identifier.exclude_rule(ExcludeRule('keyword'))
    node, start = identifier.parse('foo', 0)
    # thanks to backtracking, we get this match.
    assert node.value == 'fo'
    assert start == 2
        
def test_exclusion_3():
    identifier_initial = ExcludeRule.create('identifier-initial = ALPHA')
    no_a_initial = ExcludeRule.create('no-a-initial = "a"')
    identifier_initial.exclude_rule(no_a_initial)
    with pytest.raises(ParseError):
        identifier_initial.parse('a', 0)


def test_rule_str():
    assert str(Rule('DIGIT'))

def test_rule_from_file(tmp_path: pathlib.Path):
    grammar = ['foo = "foo"\r\n', 'bar = "bar"\r\n']
    path = tmp_path / 'test_grammar.abnf'
    path.write_text(''.join(grammar))
    
    class FromFileRule(Rule):
        pass
        
    FromFileRule.from_file(path)


def test_node_str():
    node = Node('test')
    assert str(node)

def test_node_eq():
    assert Node('test') == Node('test')

def test_literal_node_children():
    assert LiteralNode('a', 0, 1).children == []

def test_literal_node_str():
    assert str(LiteralNode('a', 0, 1))

def test_literal_node_eq():
    assert LiteralNode('a', 0, 1) == LiteralNode('a', 0, 1)

def test_literal_node_neq():
    assert LiteralNode('a', 0, 1) != LiteralNode('a', 1, 1)

def test_bin_val():
    src = "b01111000"
    node = ABNFGrammarRule('bin-val').parse_all(src)
    visitor = NumValVisitor()
    parser = visitor.visit(node)
    assert parser.value == 'x'

def test_prose_val():
    class ProseRule(Rule):
        pass
        
    rule = 'test-prose-val = <blah blah>'
    with pytest.raises(GrammarError):
        ProseRule.create(rule)


class EdgeCaseRule(Rule):
    pass

EdgeCaseRule.create('repeat-a = *"a"')
EdgeCaseRule.create('repeat-repeat-a = *(*"a")')

def test_repetition():
    node, start = EdgeCaseRule('repeat-a').parse('', 0)
    assert node.value == ''
    assert start == 0
    
def test_repetition_1():
    node, start = EdgeCaseRule('repeat-repeat-a').parse('', 0)
    assert node.value == ''
    assert start == 0



"""
Collected rules from RFC 7230, Appendix B.
https://tools.ietf.org/html/rfc7230#appendix-B
"""

from abnf.parser import Rule as _Rule
from . import rfc3986
from .misc import load_grammar_rules


@load_grammar_rules(
    [
        ("URI-reference", rfc3986.Rule("URI-reference")),
        ("absolute-URI", rfc3986.Rule("absolute-URI")),
        ("authority", rfc3986.Rule("authority")),
        ("fragment", rfc3986.Rule("fragment")),
        ("path-abempty", rfc3986.Rule("path-abempty")),
        ("port", rfc3986.Rule("port")),
        ("query", rfc3986.Rule("query")),
        ("relative-part", rfc3986.Rule("relative-part")),
        ("scheme", rfc3986.Rule("scheme")),
        ("segment", rfc3986.Rule("segment")),
        ("uri-host", rfc3986.Rule("host")),
    ]
)
class Rule(_Rule):
    """Parser rules for grammar from RFC 7230."""

    grammar = [
        "BWS = OWS",
        'Connection = *( "," OWS ) connection-option *( OWS "," [ OWS connection-option ] )',
        "Content-Length = 1*DIGIT",
        "HTTP-message = start-line *( header-field CRLF ) CRLF [ message-body ]",
        "HTTP-name = %x48.54.54.50",
        'HTTP-version = HTTP-name "/" DIGIT "." DIGIT',
        'Host = uri-host [ ":" port ]',
        "OWS = *( SP / HTAB )",
        "RWS = 1*( SP / HTAB )",
        'TE = [ ( "," / t-codings ) *( OWS "," [ OWS t-codings ] ) ]',
        'Trailer = *( "," OWS ) field-name *( OWS "," [ OWS field-name ] )',
        'Transfer-Encoding = *( "," OWS ) transfer-coding *( OWS "," [ OWS transfer-coding ] )',
        # URI-reference = <URI-reference, see [RFC3986], Section 4.1>',
        'Upgrade = *( "," OWS ) protocol *( OWS "," [ OWS protocol ] )',
        'Via = *( "," OWS ) ( received-protocol RWS received-by [ RWS comment ] ) *( OWS "," [ OWS ( received-protocol RWS received-by [ RWS comment ] ) ] )',
        # absolute-URI = <absolute-URI, see [RFC3986], Section 4.3>',
        "absolute-form = absolute-URI",
        'absolute-path = 1*( "/" segment )',
        'asterisk-form = "*"',
        # authority = <authority, see [RFC3986], Section 3.2>',
        "authority-form = authority",
        "chunk = chunk-size [ chunk-ext ] CRLF chunk-data CRLF",
        "chunk-data = 1*OCTET",
        'chunk-ext = *( ";" chunk-ext-name [ "=" chunk-ext-val ] )',
        "chunk-ext-name = token",
        "chunk-ext-val = token / quoted-string",
        "chunk-size = 1*HEXDIG",
        "chunked-body = *chunk last-chunk trailer-part CRLF",
        'comment = "(" *( ctext / quoted-pair / comment ) ")"',
        "connection-option = token",
        "ctext = HTAB / SP / %x21-27 / %x2A-5B / %x5D-7E / obs-text",
        "field-content = field-vchar [ 1*( SP / HTAB ) field-vchar ]",
        "field-name = token",
        "field-value = *( field-content / obs-fold )",
        "field-vchar = VCHAR / obs-text",
        # fragment = <fragment, see [RFC3986], Section 3.5>',
        'header-field = field-name ":" OWS field-value OWS',
        'http-URI = "http://" authority path-abempty [ "?" query ] [ "#" fragment ]',
        'https-URI = "https://" authority path-abempty [ "?" query ] [ "#" fragment ]',
        'last-chunk = 1*"0" [ chunk-ext ] CRLF',
        "message-body = *OCTET",
        "method = token",
        "obs-fold = CRLF 1*( SP / HTAB )",
        "obs-text = %x80-FF",
        'origin-form = absolute-path [ "?" query ]',
        'partial-URI = relative-part [ "?" query ]',
        # path-abempty = <path-abempty, see [RFC3986], Section 3.3>',
        # port = <port, see [RFC3986], Section 3.2.3>',
        'protocol = protocol-name [ "/" protocol-version ]',
        "protocol-name = token",
        "protocol-version = token",
        "pseudonym = token",
        'qdtext = HTAB / SP / "!" / %x23-5B / %x5D-7E / obs-text',
        # query = <query, see [RFC3986], Section 3.4>',
        'quoted-pair = "\\" ( HTAB / SP / VCHAR / obs-text )',
        "quoted-string = DQUOTE *( qdtext / quoted-pair ) DQUOTE",
        'rank = ( "0" [ "." *3DIGIT ] ) / ( "1" [ "." *3"0" ] )',
        "reason-phrase = *( HTAB / SP / VCHAR / obs-text )",
        'received-by = ( uri-host [ ":" port ] ) / pseudonym',
        'received-protocol = [ protocol-name "/" ] protocol-version',
        # relative-part = <relative-part, see [RFC3986], Section 4.2>',
        "request-line = method SP request-target SP HTTP-version CRLF",
        "request-target = origin-form / absolute-form / authority-form / asterisk-form",
        # scheme = <scheme, see [RFC3986], Section 3.1>',
        # segment = <segment, see [RFC3986], Section 3.3>',
        "start-line = request-line / status-line",
        "status-code = 3DIGIT",
        "status-line = HTTP-version SP status-code SP reason-phrase CRLF",
        't-codings = "trailers" / ( transfer-coding [ t-ranking ] )',
        't-ranking = OWS ";" OWS "q=" rank',
        'tchar = "!" / "#" / "$" / "%" / "&" / "\'" / "*" / "+" / "-" / "." / "^" / "_" / "`" / "|" / "~" / DIGIT / ALPHA',
        "token = 1*tchar",
        "trailer-part = *( header-field CRLF )",
        'transfer-coding = "chunked" / "compress" / "deflate" / "gzip" / transfer-extension',
        'transfer-extension = token *( OWS ";" OWS transfer-parameter )',
        'transfer-parameter = token BWS "=" BWS ( token / quoted-string )',
        #'uri-host = host' #<host, see [RFC3986], Section 3.2.2>',
    ]



"""
Collected rules from RFC 5322
https://tools.ietf.org/html/rfc5322
"""

from abnf.parser import Rule as _Rule
from .misc import load_grammar_rules


@load_grammar_rules()
class Rule(_Rule):
    """Rule objects generated from ABNF in RFC 5322."""

    grammar = [
        'quoted-pair = ("\\" (VCHAR / WSP)) / obs-qp',
        "FWS = ([*WSP CRLF] 1*WSP) / obs-FWS ",
        "ctext = %d33-39 / %d42-91 / %d93-126 / obs-ctext",
        "ccontent = ctext / quoted-pair / comment",
        'comment = "(" *([FWS] ccontent) [FWS] ")"',
        "CFWS = (1*([FWS] comment) [FWS]) / FWS",
        'atext = ALPHA / DIGIT / "!" / "#" / "$" / "%" / "&" / "\'" / "*" / "+" / "-" / "/" / "=" / "?" / "^" / "_" / "`" / "{" / "|" / "}" / "~"',
        "atom = [CFWS] 1*atext [CFWS]",
        'dot-atom-text = 1*atext *("." 1*atext)',
        "dot-atom = [CFWS] dot-atom-text [CFWS]",
        'specials = "(" / ")" / "<" / ">" / "[" / "]" / ":" / ";" / "@" / "\\" / "," / "." / DQUOTE',
        "qtext = %d33 / %d35-91 / %d93-126 / obs-qtext",
        "qcontent = qtext / quoted-pair",
        "quoted-string = [CFWS] DQUOTE *([FWS] qcontent) [FWS] DQUOTE [CFWS]",
        "word = atom / quoted-string",
        "phrase = 1*word / obs-phrase",
        "unstructured = (*([FWS] VCHAR) *WSP) / obs-unstruct",
        'date-time = [ day-of-week "," ] date time [CFWS]',
        "day-of-week = ([FWS] day-name) / obs-day-of-week",
        'day-name = "Mon" / "Tue" / "Wed" / "Thu" / "Fri" / "Sat" / "Sun"',
        "date = day month year",
        "day = ([FWS] 1*2DIGIT FWS) / obs-day",
        'month = "Jan" / "Feb" / "Mar" / "Apr" / "May" / "Jun" / "Jul" / "Aug" / "Sep" / "Oct" / "Nov" / "Dec"',
        "year = (FWS 4*DIGIT FWS) / obs-year",
        "time = time-of-day zone",
        'time-of-day = hour ":" minute [ ":" second ]',
        "hour = 2DIGIT / obs-hour",
        "minute = 2DIGIT / obs-minute",
        "second = 2DIGIT / obs-second",
        'zone = (FWS ( "+" / "-" ) 4DIGIT) / obs-zone',
        "address = mailbox / group",
        "mailbox = name-addr / addr-spec",
        "name-addr = [display-name] angle-addr",
        'angle-addr = [CFWS] "<" addr-spec ">" [CFWS] / obs-angle-addr',
        'group = display-name ":" [group-list] ";" [CFWS]',
        "display-name = phrase",
        'mailbox-list = (mailbox *("," mailbox)) / obs-mbox-list',
        'address-list = (address *("," address)) / obs-addr-list',
        "group-list = mailbox-list / CFWS / obs-group-list",
        'addr-spec = local-part "@" domain',
        "local-part = dot-atom / quoted-string / obs-local-part",
        "domain = dot-atom / domain-literal / obs-domain",
        'domain-literal = [CFWS] "[" *([FWS] dtext) [FWS] "]" [CFWS]',
        "dtext = %d33-90 / %d94-126 / obs-dtext ",
        "message = (fields / obs-fields) [CRLF body]",
        "body = (*(*998text CRLF) *998text) / obs-body",
        "text = %d1-9 / %d11 / %d12 / %d14-127",
        "fields = *(trace *optional-field / *(resent-date / resent-from / resent-sender / resent-to / resent-cc / resent-bcc / resent-msg-id)) *(orig-date / from / sender / reply-to / to / cc / bcc / message-id / in-reply-to / references / subject / comments / keywords / optional-field)",
        'orig-date = "Date:" date-time CRLF',
        'from = "From:" mailbox-list CRLF',
        'sender = "Sender:" mailbox CRLF',
        'reply-to = "Reply-To:" address-list CRLF',
        'to = "To:" address-list CRLF',
        'cc = "Cc:" address-list CRLF',
        'bcc = "Bcc:" [address-list / CFWS] CRLF',
        'message-id = "Message-ID:" msg-id CRLF',
        'in-reply-to = "In-Reply-To:" 1*msg-id CRLF',
        'references = "References:" 1*msg-id CRLF',
        'msg-id = [CFWS] "<" id-left "@" id-right ">" [CFWS]',
        "id-left = dot-atom-text / obs-id-left",
        "id-right = dot-atom-text / no-fold-literal / obs-id-right",
        'no-fold-literal = "[" *dtext "]"',
        'subject = "Subject:" unstructured CRLF',
        'comments = "Comments:" unstructured CRLF',
        'keywords = "Keywords:" phrase *("," phrase) CRLF',
        'resent-date = "Resent-Date:" date-time CRLF',
        'resent-from = "Resent-From:" mailbox-list CRLF',
        'resent-sender = "Resent-Sender:" mailbox CRLF',
        'resent-to = "Resent-To:" address-list CRLF',
        'resent-cc = "Resent-Cc:" address-list CRLF',
        'resent-bcc = "Resent-Bcc:" [address-list / CFWS] CRLF',
        'resent-msg-id = "Resent-Message-ID:" msg-id CRLF',
        "trace = [return] 1*received",
        'return = "Return-Path:" path CRLF',
        'path = angle-addr / ([CFWS] "<" [CFWS] ">" [CFWS])',
        'received = "Received:" *received-token ";" date-time CRLF',
        "received-token = word / angle-addr / addr-spec / domain",
        'optional-field = field-name ":" unstructured CRLF',
        "field-name = 1*ftext",
        "ftext = %d33-57 / %d59-126 ",
        "obs-NO-WS-CTL = %d1-8 / %d11 / %d12 / %d14-31 / %d127 ",
        "obs-ctext = obs-NO-WS-CTL",
        "obs-qtext = obs-NO-WS-CTL",
        "obs-utext = %d0 / obs-NO-WS-CTL / VCHAR",
        'obs-qp = "\\" (%d0 / obs-NO-WS-CTL / LF / CR)',
        "obs-body = *((*LF *CR *((%d0 / text) *LF *CR)) / CRLF)",
        # See https://www.rfc-editor.org/errata/eid1905.
        #"obs-unstruct = *((*LF *CR *(obs-utext *LF *CR)) / FWS)",
        "obs-unstruct = *( (*CR 1*(obs-utext / FWS)) / 1*LF ) *CR",
        'obs-phrase = word *(word / "." / CFWS)',
        'obs-phrase-list = [phrase / CFWS] *("," [phrase / CFWS])',
        "obs-FWS = 1*WSP *(CRLF 1*WSP)",
        "obs-day-of-week = [CFWS] day-name [CFWS]",
        "obs-day = [CFWS] 1*2DIGIT [CFWS]",
        "obs-year = [CFWS] 2*DIGIT [CFWS]",
        "obs-hour = [CFWS] 2DIGIT [CFWS]",
        "obs-minute = [CFWS] 2DIGIT [CFWS]",
        "obs-second = [CFWS] 2DIGIT [CFWS]",
        'obs-zone = "UT" / "GMT" / "EST" / "EDT" / "CST" / "CDT" / "MST" / "MDT" / "PST" / "PDT" / %d65-73 / %d75-90 / %d97-105 / %d107-122 ',
        'obs-angle-addr = [CFWS] "<" obs-route addr-spec ">" [CFWS]',
        'obs-route = obs-domain-list ":"',
        'obs-domain-list = *(CFWS / ",") "@" domain *("," [CFWS] ["@" domain])',
        'obs-mbox-list = *([CFWS] ",") mailbox *("," [mailbox / CFWS])',
        'obs-addr-list = *([CFWS] ",") address *("," [address / CFWS])',
        'obs-group-list = 1*([CFWS] ",") [CFWS]',
        'obs-local-part = word *("." word)',
        'obs-domain = atom *("." atom)',
        "obs-dtext = obs-NO-WS-CTL / quoted-pair",
        "obs-fields = *(obs-return / obs-received / obs-orig-date / obs-from / obs-sender / obs-reply-to / obs-to / obs-cc / obs-bcc / obs-message-id / obs-in-reply-to / obs-references / obs-subject / obs-comments / obs-keywords / obs-resent-date / obs-resent-from / obs-resent-send / obs-resent-rply / obs-resent-to / obs-resent-cc / obs-resent-bcc / obs-resent-mid / obs-optional)",
        'obs-orig-date = "Date" *WSP ":" date-time CRLF',
        'obs-from = "From" *WSP ":" mailbox-list CRLF',
        'obs-sender = "Sender" *WSP ":" mailbox CRLF',
        'obs-reply-to = "Reply-To" *WSP ":" address-list CRLF',
        'obs-to = "To" *WSP ":" address-list CRLF',
        'obs-cc = "Cc" *WSP ":" address-list CRLF',
        'obs-bcc = "Bcc" *WSP ":" (address-list / (*([CFWS] ",") [CFWS])) CRLF',
        'obs-message-id = "Message-ID" *WSP ":" msg-id CRLF',
        'obs-in-reply-to = "In-Reply-To" *WSP ":" *(phrase / msg-id) CRLF',
        'obs-references = "References" *WSP ":" *(phrase / msg-id) CRLF',
        "obs-id-left = local-part",
        "obs-id-right = domain",
        'obs-subject = "Subject" *WSP ":" unstructured CRLF',
        'obs-comments = "Comments" *WSP ":" unstructured CRLF',
        'obs-keywords = "Keywords" *WSP ":" obs-phrase-list CRLF',
        'obs-resent-from = "Resent-From" *WSP ":" mailbox-list CRLF',
        'obs-resent-send = "Resent-Sender" *WSP ":" mailbox CRLF',
        'obs-resent-date = "Resent-Date" *WSP ":" date-time CRLF',
        'obs-resent-to = "Resent-To" *WSP ":" address-list CRLF',
        'obs-resent-cc = "Resent-Cc" *WSP ":" address-list CRLF',
        'obs-resent-bcc = "Resent-Bcc" *WSP ":" (address-list / (*([CFWS] ",") [CFWS])) CRLF',
        'obs-resent-mid = "Resent-Message-ID" *WSP ":" msg-id CRLF',
        'obs-resent-rply = "Resent-Reply-To" *WSP ":" address-list CRLF',
        'obs-return = "Return-Path" *WSP ":" path CRLF',
        'obs-received = "Received" *WSP ":" *received-token CRLF',
        'obs-optional = field-name *WSP ":" unstructured CRLF',
    ]


#Rule('unstructured').first_match_alternation = True



#add tests for comment, specials, obs-qp for backslash.

import pytest

from abnf.grammars import rfc5322
from abnf.parser import ParseError

@pytest.mark.parametrize("src", [
    r'simple@example.com', 
    r'very.common@example.com', 
    r'disposable.style.email.with+symbol@example.com', 
    r'other.email-with-hyphen@example.com', 
    r'fully-qualified-domain@example.com', 
    r'user.name+tag+sorting@example.com',
    r'x@example.com',
    r'example-indeed@strange-example.com',
    r'admin@mailserver1',
    r'example@s.example',
    r'" "@example.org',
    r'"john..doe"@example.org',
    r'mailhost!username@example.org',
    r'user%example.com@example.org',
    ])
def test_address(src: str):
    rfc5322.Rule('address').parse_all(src)

@pytest.mark.parametrize("src", [
    r'Abc.example.com', # (no @ character)
    r'A@b@c@example.com', # (only one @ is allowed outside quotation marks)
    r'a"b(c)d,e:f;g<h>i[j\k]l@example.com', # (none of the special characters in this local-part are allowed outside quotation marks)
    r'just"not"right@example.com', # (quoted strings must be dot separated or the only element making up the local-part)
    r'this is"not\allowed@example.com', # (spaces, quotes, and backslashes may only exist when within quoted strings and preceded by a backslash)
    r'this\ still\"not\\allowed@example.com', # (even if escaped (preceded by a backslash), spaces, quotes, and backslashes must still be contained by quotes)
    ])    
def test_address_bad(src: str):
    with pytest.raises(ParseError):
        rfc5322.Rule('address').parse_all(src)

@pytest.mark.parametrize("src", [
'''From: John Doe <jdoe@machine.example>\r\nTo: Mary Smith <mary@example.net>\r\nSubject: Saying Hello\r\nDate: Fri, 21 Nov 1997 09:55:06 -0600\r\nMessage-ID: <1234@local.machine.example>\r\n\r\n
This is a message just to say hello.
So, "Hello".''',


]
)
def test_parse_example(src: str):
    rfc5322.Rule('message').parse_all(src)




"""Parser generator for ABNF grammars."""

try:
    from importlib.metadata import PackageNotFoundError, metadata  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import metadata, PackageNotFoundError  # type: ignore

from .parser import GrammarError, LiteralNode, Node, NodeVisitor, ParseError, Rule

__all__ = [
    "Rule",
    "Node",
    "LiteralNode",
    "NodeVisitor",
    "ParseError",
    "GrammarError",
    "__version__",
]

try:
    __version__ = metadata(__name__)["version"]
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = ""



"""
Collected rules from RFC 7234
https://tools.ietf.org/html/rfc7234
"""

from abnf.parser import Rule as _Rule
from . import rfc7230, rfc7231
from .misc import load_grammar_rules


@load_grammar_rules(
    [
        ("OWS", rfc7230.Rule("OWS")),
        ("field-name", rfc7230.Rule("field-name")),
        ("quoted-string", rfc7230.Rule("quoted-string")),
        ("token", rfc7230.Rule("token")),
        ("port", rfc7230.Rule("port")),
        ("pseudonym", rfc7230.Rule("pseudonym")),
        ("uri-host", rfc7230.Rule("uri-host")),
        ("HTTP-date", rfc7231.Rule("HTTP-date")),
    ]
)
class Rule(_Rule):
    """Rules from RFC 7234."""

    grammar = [
        "Age = delta-seconds",
        'Cache-Control = *( "," OWS ) cache-directive *( OWS "," [ OWS cache-directive ] )',
        "Expires = HTTP-date",
        # HTTP-date = <HTTP-date, see [RFC7231], Section 7.1.1.1>,
        # OWS = <OWS, see [RFC7230], Section 3.2.3>
        'Pragma = *( "," OWS ) pragma-directive *( OWS "," [ OWS pragma-directive ] )',
        'Warning = *( "," OWS ) warning-value *( OWS "," [ OWS warning-value ] )',
        'cache-directive = token [ "=" ( token / quoted-string ) ]',
        "delta-seconds = 1*DIGIT",
        'extension-pragma = token [ "=" ( token / quoted-string ) ]',
        # field-name = <field-name, see [RFC7230], Section 3.2>
        # port = <port, see [RFC7230], Section 2.7>
        'pragma-directive = "no-cache" / extension-pragma',
        # pseudonym = <pseudonym, see [RFC7230], Section 5.7.1>
        # quoted-string = <quoted-string, see [RFC7230], Section 3.2.6>
        # token = <token, see [RFC7230], Section 3.2.6>
        # uri-host = <uri-host, see [RFC7230], Section 2.7>
        'warn-agent = ( uri-host [ ":" port ] ) / pseudonym',
        "warn-code = 3DIGIT",
        "warn-date = DQUOTE HTTP-date DQUOTE",
        "warn-text = quoted-string",
        "warning-value = warn-code SP warn-agent SP warn-text [ SP warn-date ]",
    ]
